{-------------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: RequestProcessor.pas, released 11 Mar 2001.

The Initial Developer of the Original Code is: Jrgen Gntherodt
(jguentherodt@users.sourceforge.net).

Portions created by Jrgen Gntherodt are Copyright (C) 2001 Jrgen Gntherodt.
All Rights Reserved.

Alternatively, the contents of this file may be used under the terms of the
GNU General Public License Version 2 or later (the "GPL"), in which case
the provisions of the GPL are applicable instead of those above.
If you wish to allow use of your version of this file only under the terms
of the GPL and not to allow others to use your version of this file
under the MPL, indicate your decision by deleting the provisions above and
replace them with the notice and other provisions required by the GPL.
If you do not delete the provisions above, a recipient may use your version
of this file under either the MPL or the GPL.

$Id: RequestProcessor.pas,v 1.3 2001/03/17 04:34:59 jguentherodt Exp $

You may retrieve the latest version of this file at the Open Perl IDE
home page, located at http://www.lost-sunglasses.de
-------------------------------------------------------------------------------}
unit RequestProcessor;

interface

uses
  message_interface, Classes, SysUtils, syncObjs, windows;

type
  TRequest = class
  private
    m_nMessageID: Integer;
    m_sData: String;
  public
    constructor Create(nMessageID: Integer; sData: String);
    property MessageID: Integer Read m_nMessageID;
    property Data: String Read m_sData;
  end;

  TRequestProcessor = class(TThread)
  private
    m_Dispatcher: IMessageDispatcher;
    m_slMessageQueue: TStringList;
    m_csLock: TCriticalSection;
    m_nMessageID: Integer;
    m_sStdOutData: String;
    m_sStdErrData: String;
    m_sMessageData: String;
    m_nStdOutResult: Integer;
    m_nStdErrResult: Integer;
    m_nTerminateResult: Integer;
    m_nMessageResult: Integer;
    m_OriginalMessageReceiver: IMessageReceiver;
    m_evQueueNotEmpty: TEvent;
    m_evProcessingPaused: TEvent;
    m_evPauseProcessing: TEvent;
    m_evContinueProcessing: TEvent;
    procedure Lock;
    procedure Unlock;
    procedure ClearMessageQueue;
    function GetNextMessage(var nMessageID: Integer; var sData: String): Boolean;
    function HandleReceivedMessage(nMsgID: Integer; sData: String): Integer;
    function ReceiveFromStdOut(sData: String): Integer;
    function ReceiveFromStdErr(sData: String): Integer;
    function HandleTerminate: Integer;
    // Synchronized functions
    procedure SyncHandleReceivedMessage;
    procedure SyncReceiveFromStdOut;
    procedure SyncReceiveFromStdErr;
    procedure SyncHandleTerminate;
    procedure DispatchMessage(nMessageID: Integer; sData: String);
  protected
    procedure Execute; override;
  public
    constructor Create(Dispatcher: IMessageDispatcher);
    procedure PauseProcessing;
    procedure ContinueProcessing;
    function Request(nMessageID: Integer; sData: String; nPriority: Integer): Integer;
    property Dispatcher: IMessageDispatcher Read m_Dispatcher;
    property ProcessingPaused: TEvent Read m_evProcessingPaused;
  end;

  TMessageSynchronizer = class(TInterfacedObject, IMessageReceiver)
  private
    m_RequestProcessor: TRequestProcessor;
  protected
    function HandleMessage(nMsgID: Integer; sData: String): Integer;
    function ReceiveFromStdOut(sData: String): Integer;
    function ReceiveFromStdErr(sData: String): Integer;
    function HandleTerminate: Integer;
  public
    constructor Create(RequestProcessor: TRequestProcessor);
  end;


implementation

////////////////////////////////////////////////////////////////////////////////
//  TMessageSynchronizer = class(TInterfacedObject, IMessageReceiver)
////////////////////////////////////////////////////////////////////////////////
constructor TMessageSynchronizer.Create(RequestProcessor: TRequestProcessor);
begin
  inherited Create;
  m_RequestProcessor := RequestProcessor;
end;


function TMessageSynchronizer.HandleMessage(nMsgID: Integer; sData: String): Integer;
begin
  Result := m_RequestProcessor.HandleReceivedMessage(nMsgID, sData);
end;


function TMessageSynchronizer.ReceiveFromStdOut(sData: String): Integer;
begin
  Result := m_RequestProcessor.ReceiveFromStdOut(sData);
end;


function TMessageSynchronizer.ReceiveFromStdErr(sData: String): Integer;
begin
  Result := m_RequestProcessor.ReceiveFromStdErr(sData);
end;


function TMessageSynchronizer.HandleTerminate: Integer;
begin
  Result := m_RequestProcessor.HandleTerminate;
end;


////////////////////////////////////////////////////////////////////////////////
//  TRequest = class
////////////////////////////////////////////////////////////////////////////////
constructor TRequest.Create(nMessageID: Integer; sData: String);
begin
  inherited Create;
  m_nMessageID := nMessageID;
  m_sData := sData
end;


////////////////////////////////////////////////////////////////////////////////
//  TRequestProcessor = class
////////////////////////////////////////////////////////////////////////////////
constructor TRequestProcessor.Create(Dispatcher: IMessageDispatcher);
begin
  m_Dispatcher := Dispatcher;
  if m_Dispatcher <> nil then begin
    m_OriginalMessageReceiver := Dispatcher.OnReceiveMessage;
    Dispatcher.OnReceiveMessage := TMessageSynchronizer.Create(Self) as IMessageReceiver;
  end;
  m_slMessageQueue := TStringList.Create;
  m_slMessageQueue.Sorted := True;
  m_slMessageQueue.Duplicates := dupIgnore;
  m_csLock := TCriticalSection.Create;
  m_evQueueNotEmpty := TEvent.Create(nil, True, False, '');
  m_evProcessingPaused := TEvent.Create(nil, True, False, '');
  m_evPauseProcessing := TEvent.Create(nil, False, False, '');
  m_evContinueProcessing := TEvent.Create(nil, False, False, '');
  inherited Create(False);
end;


procedure TRequestProcessor.PauseProcessing;
begin
  m_evPauseProcessing.SetEvent;
end;


procedure TRequestProcessor.ContinueProcessing;
begin
  m_evContinueProcessing.SetEvent;
end;


procedure TRequestProcessor.DispatchMessage(nMessageID: Integer; sData: String);
begin
  if m_Dispatcher <> nil
    then m_Dispatcher.SendMessage(nMessageID, sData);
end;


procedure TRequestProcessor.SyncReceiveFromStdOut;
begin
  if m_OriginalMessageReceiver <> nil
    then m_nStdOutResult := m_OriginalMessageReceiver.ReceiveFromStdOut(m_sStdOutData)
    else m_nStdOutResult := 0;
end;


procedure TRequestProcessor.SyncReceiveFromStdErr;
begin
  if m_OriginalMessageReceiver <> nil
    then m_nStdErrResult := m_OriginalMessageReceiver.ReceiveFromStdErr(m_sStdErrData)
    else m_nStdErrResult := 0;
end;


procedure TRequestProcessor.SyncHandleTerminate;
begin
  if m_OriginalMessageReceiver <> nil
    then m_nTerminateResult := m_OriginalMessageReceiver.HandleTerminate
    else m_nTerminateResult := 0;
end;


procedure TRequestProcessor.SyncHandleReceivedMessage;
begin
  if m_OriginalMessageReceiver <> nil
    then m_nMessageResult := m_OriginalMessageReceiver.HandleMessage(m_nMessageID, m_sMessageData)
    else m_nMessageResult := 0;
end;


function TRequestProcessor.ReceiveFromStdOut(sData: String): Integer;
begin
  m_sStdOutData := sData;
  Synchronize(SyncReceiveFromStdOut);
  Result := m_nStdOutResult;
end;


function TRequestProcessor.ReceiveFromStdErr(sData: String): Integer;
begin
  m_sStdErrData := sData;
  Synchronize(SyncReceiveFromStdErr);
  Result := m_nStdErrResult;
end;


function TRequestProcessor.HandleTerminate: Integer;
begin
  Synchronize(SyncHandleTerminate);
  Result := m_nTerminateResult;
end;


function TRequestProcessor.HandleReceivedMessage(nMsgID: Integer; sData: String): Integer;
begin
  m_nMessageID := nMsgID;
  m_sMessageData := sData;
  Synchronize(SyncHandleReceivedMessage);
  Result := m_nMessageResult;
end;


function TRequestProcessor.GetNextMessage(var nMessageID: Integer; var sData: String): Boolean;
var
  req: TRequest;
begin
  Lock;
  if m_slMessageQueue.Count > 0 then begin
    req := TRequest(m_slMessageQueue.Objects[0]);
    m_slMessageQueue.Delete(0);
    if m_slMessageQueue.Count = 0 then m_evQueueNotEmpty.ResetEvent;
  end else req := nil;
  Unlock;
  Result := req <> nil;
  if Result then begin
    nMessageID := req.MessageID;
    sData := req.Data;
    req.Free;
  end;
end;


procedure TRequestProcessor.Execute;
var
  nMessageID: Integer;
  sData: String;
  handles: TWOHandleArray;
begin
  handles[0] := m_evPauseProcessing.Handle;
  handles[1] := m_evQueueNotEmpty.Handle;
  while not Terminated do begin
    case WaitForMultipleObjects(2, @handles, False, 500) of
      WAIT_OBJECT_0: begin
        m_evProcessingPaused.SetEvent;
        WaitForSingleObject(m_evContinueProcessing.Handle, INFINITE);
        m_evProcessingPaused.ResetEvent;
      end;
      WAIT_OBJECT_0 + 1: begin
        if GetNextMessage(nMessageID, sData) then DispatchMessage(nMessageID, sData);
      end;
    else
    end;
  end;
  ClearMessageQueue;
  m_slMessageQueue.Free;
  m_csLock.Free;
  m_evQueueNotEmpty.Free;
  m_evProcessingPaused.Free;
  m_evPauseProcessing.Free;
  if m_Dispatcher <> nil then m_Dispatcher.OnReceiveMessage := m_OriginalMessageReceiver;
end;


procedure TRequestProcessor.Lock;
begin
 m_csLock.Enter;
end;


procedure TRequestProcessor.Unlock;
begin
  m_csLock.Leave;
end;


procedure TRequestProcessor.ClearMessageQueue;
var
  i: Integer;
begin
  Lock;
  for i := 0 to m_slMessageQueue.Count - 1 do begin
    m_slMessageQueue.Objects[i].Free;
  end;
  m_slMessageQueue.Clear;
  Unlock;
end;


function TRequestProcessor.Request(nMessageID: Integer; sData: String;
  nPriority: Integer): Integer;
var
  reqNew: TRequest;
begin
  reqNew := TRequest.Create(nMessageID, sData);
  Lock;
  m_slMessageQueue.AddObject(IntToHex(nPriority, 8) + sData, reqNew);
  Result := m_slMessageQueue.Count;
  m_evQueueNotEmpty.SetEvent;
  Unlock;
end;


end.
