{-------------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: clProcess.pas, released 11 Mar 2001.

The Initial Developer of the Original Code is: Jrgen Gntherodt
(jguentherodt@users.sourceforge.net).

Portions created by Jrgen Gntherodt are Copyright (C) 2001 Jrgen Gntherodt.
All Rights Reserved.

Alternatively, the contents of this file may be used under the terms of the
GNU General Public License Version 2 or later (the "GPL"), in which case
the provisions of the GPL are applicable instead of those above.
If you wish to allow use of your version of this file only under the terms
of the GPL and not to allow others to use your version of this file
under the MPL, indicate your decision by deleting the provisions above and
replace them with the notice and other provisions required by the GPL.
If you do not delete the provisions above, a recipient may use your version
of this file under either the MPL or the GPL.

$Id: clProcess.pas,v 1.3 2001/03/17 04:34:59 jguentherodt Exp $

You may retrieve the latest version of this file at the Open Perl IDE
home page, located at http://www.lost-sunglasses.de
-------------------------------------------------------------------------------}
unit clProcess;

interface

uses
  Classes, windows, syncObjs;

// This class encapsulates a command line process, especially supression
// of console window and redirection of STDIN, STDOUT and STDERR.
// STDHandles are redirected, if and only if the console window is suppressed.

type
  TCLProcessEvent = procedure(Sender: TObject; sData: String; bError, bTerminate: Boolean) of object;

  TCLProcess = class(TThread)
  private
    m_sCommandLine: String;
    m_sCurrentDirectory: String;
    m_bShowConsole: Boolean;
    m_event: TCLProcessEvent;
    m_sError: String;
    m_sData: String;
    // Pipe handles
    m_hPipeOutputRead: THandle;
    m_hPipeOutputWrite: THandle;
    m_hPipeErrorRead: THandle;
    m_hPipeErrorWrite: THandle;
    m_hPipeInputRead: THandle;
    m_hPipeInputWrite: THandle;
    //
    m_pi: PROCESS_INFORMATION;
    m_hStdErrThread: THandle;
    m_csLock: TCriticalSection;
    procedure DoOutput;
    procedure DoError;
    procedure DoTerminateApp;
    procedure InternalCreateProcess;
    function GetProcessHandle: THandle;
    function StdErr: DWord;
  protected
    procedure Lock; virtual;
    procedure UnLock; virtual;
    procedure Execute; override;
  public
    constructor Create(sCommandLine, sCurrentDirectory: String; event: TCLProcessEvent);
    destructor Destroy; override;
    function Input(sData: String): Boolean;
    property CommandLine: String Read m_sCommandLine;
    property ShowConsole: Boolean Read m_bShowConsole;
    property ProcessHandle: THandle Read GetProcessHandle;
  end;

implementation


function StdErrThread(p: Pointer): DWord; stdcall;
begin
  Result := TCLProcess(p).StdErr;
end;


////////////////////////////////////////////////////////////////////////////////
//  TCLProcess = class(TThread)
////////////////////////////////////////////////////////////////////////////////
constructor TCLProcess.Create(sCommandLine, sCurrentDirectory: String; event: TCLProcessEvent);
var
  dw: DWord;
begin
  m_sCommandLine := sCommandLine;
  m_sCurrentDirectory := sCurrentDirectory;
  m_event := event;
  m_bShowConsole := not(Assigned(m_event));
  m_hPipeOutputRead := 0;
  m_hPipeOutputWrite := 0;
  m_hPipeErrorRead := 0;
  m_hPipeErrorWrite := 0;
  m_hPipeInputRead := 0;
  m_hPipeInputWrite := 0;
  m_csLock := TCriticalSection.Create;
  InternalCreateProcess;
  m_hStdErrThread := CreateThread(nil, 0, @StdErrThread, Self, 0, dw);
  inherited Create(False);
end;


destructor TCLProcess.Destroy;
begin
  CloseHandle(m_hPipeInputWrite);
  inherited Destroy;
end;


procedure TCLProcess.Lock;
begin
  m_csLock.Enter;
end;


procedure TCLProcess.UnLock;
begin
  m_csLock.Leave;
end;


function TCLProcess.StdErr: DWord;
var
  buf: array[0..1023] of char;
  dwBytesRead: DWord;
  b: Boolean;
begin
  while not Terminated do begin
    b := ReadFile(m_hPipeErrorRead, buf, 64, dwBytesRead, nil);
    if not b then break;
    buf[dwBytesRead] := #0;
    m_sError := PChar(@buf[0]);
    Lock;
    Synchronize(DoError);
    Unlock;
  end;
  CloseHandle(m_hPipeErrorRead);
  Result := 0;
end;


function TCLProcess.GetProcessHandle: THandle;
begin
  Result := m_pi.hProcess;
end;


procedure TCLProcess.InternalCreateProcess;
var
  sa: SECURITY_ATTRIBUTES;
  si: STARTUPINFO;
  pCurrentDirectory: PChar;
begin
  FillChar(sa, SizeOf(sa), 0);
  with sa do begin
    nLength := SizeOf(sa);
    sa.bInheritHandle := True;
    sa.lpSecurityDescriptor := nil;
  end;
  if not m_bShowConsole then begin
    CreatePipe(m_hPipeOutputRead, m_hPipeOutputWrite, @sa, 0);
    CreatePipe(m_hPipeErrorRead, m_hPipeErrorWrite, @sa, 0);
    CreatePipe(m_hPipeInputRead, m_hPipeInputWrite, @sa, 0);
  end;
  FillChar(si, SizeOf(si), 0);
  with si do begin
    cb := SizeOf(si);
    if m_bShowConsole then begin
      dwFlags := STARTF_USESHOWWINDOW;
      wShowWindow := SW_SHOW;
    end else begin
      dwFlags := STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
      wShowWindow := SW_HIDE;
    end;
    hStdInput := m_hPipeInputRead;
    hStdOutput := m_hPipeOutputWrite;
    hStdError := m_hPipeErrorWrite;
  end;
//  CreateProcess(nil, PChar(m_sCommandLine), @sa, @sa, True, 0,
//    PChar(m_sEnvironment), nil, si, m_pi);
  if m_sCurrentDirectory <> ''
    then pCurrentDirectory := PChar(m_sCurrentDirectory)
    else pCurrentDirectory := nil;
  CreateProcess(nil, PChar(m_sCommandLine), @sa, @sa, True, 0,
    nil, pCurrentDirectory, si, m_pi);
  CloseHandle(m_hPipeInputRead);
  CloseHandle(m_hPipeOutputWrite);
  CloseHandle(m_hPipeErrorWrite);
end;


procedure TCLProcess.DoError;
begin
  if Assigned(m_event) then m_event(Self, m_sError, True, False);
end;

procedure TCLProcess.DoOutput;
begin
  if Assigned(m_event) then m_event(Self, m_sData, False, False);
end;


procedure TCLProcess.DoTerminateApp;
begin
  if Assigned(m_event) then m_event(Self, '', False, True);
end;


procedure TCLProcess.Execute;
var
  buf: array[0..1023] of char;
  dwBytesRead: DWord;
  b: Boolean;
begin
  while not Terminated do begin
    b := ReadFile(m_hPipeOutputRead, buf, 64, dwBytesRead, nil);
    if not b then break;
    buf[dwBytesRead] := #0;
    m_sData := PChar(@buf[0]);
    Lock;
    Synchronize(DoOutput);
    Unlock;
  end;
  CloseHandle(m_pi.hProcess);
  CloseHandle(m_hPipeOutputRead);
  WaitForSingleObject(m_hStdErrThread, INFINITE);
  Lock;
  Synchronize(DoTerminateApp);
  Unlock;
  m_csLock.Free;
end;


function TCLProcess.Input(sData: String): Boolean;
var
  dwBytesWritten: DWord;
begin
  Result := WriteFile(m_hPipeInputWrite, sData[1], Length(sData), dwBytesWritten, nil);
end;

end.
